# This module is part of GitPython and is released under the
# 3-Clause BSD License: https://opensource.org/license/bsd-3-clause/

"""Initialize the index package."""

from .base import *  # noqa: F401 F403
from .typ import *  # noqa: F401 F403
